(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Edg imports Trs Substitution DpFramework begin

datatype ('f,'v) hterm = Hole ("\<box>") | HFun 'f "('f,'v) hterm list"

instantiation "hterm" :: (sshow,type) sshow
begin
fun sshow_hterm 
  where "sshow_hterm Hole = ''_''"
      | "sshow_hterm (HFun f ts) = (sshow f +++   
            (case ts of Nil \<Rightarrow> sEmpty
               | (Cons s ss) \<Rightarrow>  (''('' +++  (sshow_hterm s +++ (sConcat (map (\<lambda> x. 
sConcat [ '','', sshow_hterm x]) ss)) +++ '')''))))"
instance ..
end


fun equiv_class :: "('f,'v) hterm \<Rightarrow> ('f,'v) term set" ("[_]")
  where "[\<box>] = UNIV"
     |  "[HFun f hs] = {Fun f ts | ts. length ts = length hs \<and> (\<forall> i < length hs. ts ! i \<in> [hs ! i])}"

lemma equiv_classes_nonempty : fixes h :: "('f,'v) hterm" shows "\<exists> t. t \<in> [h]" 
proof -
  let ?P = "\<lambda> hs. (\<exists> ts. length hs = length ts \<and> (\<forall> i < length hs. ts ! i \<in> [hs ! i]))"
  show ?thesis 
  proof (induct h)
    show "\<exists> t. t \<in> [\<box>]" by auto
  next
    fix f :: 'f and hs :: "('f,'v) hterm list"
    assume "?P hs"
    from this obtain ts where "Fun f ts \<in> [HFun f hs]" by force 
    thus "\<exists> t. t \<in> [HFun f hs]" ..
  next
    show "?P []" by simp
  next
    fix h :: "('f,'v) hterm" and hs :: "('f,'v) hterm list"
    assume et: "\<exists> t. t \<in> [h]" and P: "?P hs"
    from et obtain t where t: "t \<in> [h]" ..
    from P obtain ts where ts: "length hs = length ts \<and> (\<forall> i < length hs. ts ! i \<in> [hs ! i])" ..
    hence len: "length (h # hs) = length (t # ts)" by simp
    have "\<forall> i < length (h # hs). (t # ts) ! i \<in> [(h # hs) ! i]" 
    proof (rule allI, rule impI)
      fix i
      assume len2: "i < length (h # hs)"
      show "(t # ts) ! i \<in> [(h # hs) ! i]"
      proof (cases i)
	case 0 with t show ?thesis by simp
      next
	case (Suc j)
	with len len2 have "j < length hs" by simp
	with Suc ts show ?thesis by simp
      qed
    qed
    with len show "?P (h # hs)" by force
  qed
qed
  

definition match :: "('f,'v) hterm \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  where "match h t == (\<exists> \<sigma>. t \<cdot> \<sigma> \<in> [h])"


(* an executable version of match which is not yet fully implemented *)

fun mergeHTerms :: "('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list option"
  where "mergeHTerms [] [] = Some []"
    |   "mergeHTerms [] (_#_) = None"
    |   "mergeHTerms (_#_) [] = None"
    |   "mergeHTerms (\<box> # hsl) (h # hsr) = (case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> Some (h # res))"
    |   "mergeHTerms (HFun f hs # hsl) (\<box> # hsr) = (case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> Some (HFun f hs # res))" 
    |   "mergeHTerms (HFun f fs # hsl) (HFun g gs # hsr) = 
            (if f = g then (case mergeHTerms fs gs of None \<Rightarrow> None | Some mhs \<Rightarrow> (case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> 
                             Some (HFun f mhs # res))) else None)"


  
lemma mergeHTermsComplete : "mergeHTerms hsl hsr = None \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists> i < length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
proof (induct rule: mergeHTerms.induct)
  case 1 thus ?case by simp
next
  case 2 thus ?case by simp
next
  case 3 thus ?case by simp
next
  fix hsl h hsr 
  assume rec: "mergeHTerms hsl hsr = None \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
  assume res: "mergeHTerms (\<box> # hsl) (h # hsr) = None"
  show "length (\<box> # hsl) \<noteq> length (h # hsr) \<or> (\<exists>i<length (\<box> # hsl). [(\<box> # hsl) ! i] \<inter> [(h # hsr) ! i] \<subseteq> {})" 
  proof (cases "mergeHTerms hsl hsr")
    case None with rec show ?thesis by force
  next
    case (Some res)
    with res show ?thesis by simp
  qed
next
  fix f :: 'f and hs hsl hsr :: "('f,'v) hterm list"
  let ?h = "HFun f hs"
  assume rec: "mergeHTerms hsl hsr = None \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
  assume res: "mergeHTerms (?h # hsl) (\<box> # hsr) = None"
  show "length (?h # hsl) \<noteq> length (\<box> # hsr) \<or> (\<exists>i<length (?h # hsl). [(?h # hsl) ! i] \<inter> [(\<box> # hsr) ! i] \<subseteq> {})" 
  proof (cases "mergeHTerms hsl hsr")
    case None with rec show ?thesis by force
  next
    case (Some res)
    with res show ?thesis by simp
  qed
next
  fix f fs hsl g gs hsr
  assume rec1: "\<lbrakk>f = g; mergeHTerms fs gs = None\<rbrakk> \<Longrightarrow> length fs \<noteq> length gs \<or> (\<exists>i<length fs. [fs ! i] \<inter> [gs ! i] \<subseteq> {})"
   and rec2: "\<And>a. \<lbrakk>f = g; mergeHTerms fs gs = Some a; mergeHTerms hsl hsr = None\<rbrakk>
            \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
   and noneRes: "mergeHTerms (HFun f fs # hsl) (HFun g gs # hsr) = None"
  show "length (HFun f fs # hsl) \<noteq> length (HFun g gs # hsr) \<or>
    (\<exists>i<length (HFun f fs # hsl). [(HFun f fs # hsl) ! i] \<inter> [(HFun g gs # hsr) ! i] \<subseteq> {})"
  proof (cases "f = g")
    case False
    have "[HFun f fs] \<inter> [HFun g gs] \<subseteq> {}" 
    proof 
      fix t
      assume "t \<in> [HFun f fs] \<inter> [HFun g gs]"
      with False have False by (cases t, auto) 
      thus "t \<in> {}" ..
    qed
    hence nearly: "[(HFun f fs # hsl) ! 0] \<inter> [(HFun g gs # hsr) ! 0] \<subseteq> {}" by simp
    have "0 < length (HFun f fs # hsl)" by simp
    with nearly show ?thesis by blast 
  next
    case True
    show ?thesis 
    proof (cases "mergeHTerms fs gs")
      case None
      with `f = g` and rec1 have prop1: "length fs \<noteq> length gs \<or> (\<exists> i < length fs. [fs ! i] \<inter> [gs ! i] \<subseteq> {})" by simp
      thus ?thesis
      proof
	assume len: "length fs \<noteq> length gs" 
	have "[HFun f fs] \<inter> [HFun g gs] \<subseteq> {}"
	proof
	  fix t
	  assume "t \<in> [HFun f fs] \<inter> [HFun g gs]"
	  with len have False by (cases t, auto)
	  thus "t \<in> {}" ..
	qed
	hence nearly: "[(HFun f fs # hsl) ! 0] \<inter> [(HFun g gs # hsr) ! 0] \<subseteq> {}" by simp
	have "0 < length (HFun f fs # hsl)" by simp
	with nearly show ?thesis by blast
      next	
	assume "\<exists>i<length fs. [fs ! i] \<inter> [gs ! i] \<subseteq> {}"
	from this obtain i where len: "i < length fs" and ne: "[fs ! i] \<inter> [gs ! i] \<subseteq> {}" by force
	have "[HFun f fs] \<inter> [HFun g gs] \<subseteq> {}"
	proof
	  fix t
	  assume "t \<in> [HFun f fs] \<inter> [HFun g gs]"
	  with len and ne have False by (cases t, auto)
	  thus "t \<in> {}" ..
	qed
	hence nearly: "[(HFun f fs # hsl) ! 0] \<inter> [(HFun g gs # hsr) ! 0] \<subseteq> {}" by simp
	have "0 < length (HFun f fs # hsl)" by simp
	with nearly show ?thesis by blast
      qed
    next
      case (Some res)
      with `f = g` and rec2 and noneRes have 
	"length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})" by force
      thus ?thesis
      proof
	assume "length hsl \<noteq> length hsr" thus ?thesis by simp
      next
	assume "\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {}" thus ?thesis by force
      qed
    qed
  qed
qed

lemma  fixes P :: "('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list  \<Rightarrow> bool"
  assumes EE: "P [] [] []" 
  and BL: "\<And>h hsl hsr mhs. \<lbrakk>P hsl hsr mhs\<rbrakk> \<Longrightarrow> P (\<box>  # hsl) (h # hsr) (h # mhs)"
  and BR: "\<And>h hsl hsr mhs. \<lbrakk>P hsl hsr mhs\<rbrakk> \<Longrightarrow> P (h  # hsl) (\<box> # hsr) (h # mhs)"
  and FF: "\<And>f ls rs hsl hsr m mhs. \<lbrakk>P ls rs m; P hsl hsr mhs\<rbrakk> \<Longrightarrow> P (HFun f ls # hsl) (HFun f rs # hsr) (HFun f m # mhs)"
  and Pre: "mergeHTerms hsl hsr = Some mhs"
  shows mergeHTerms_induct [case_names EE BL BR FF Pre] : "P hsl hsr mhs"
using Pre
proof (induct arbitrary: mhs rule: mergeHTerms.induct) 
  case 1 with EE show ?case by simp
next
  case 2 thus ?case by simp
next
  case (3 h hls mhs) thus ?case by simp
next
  case (4 hsl h hsr mhs) with BL show ?case by (cases "mergeHTerms hsl hsr", auto)
next
  case (5 f hs hsl hsr mhs) with BR show ?case by (cases "mergeHTerms hsl hsr", auto)
next
  case (6 f fs hsl g gs hsr mhs)
  from `mergeHTerms (HFun f fs # hsl) (HFun g gs # hsr) = Some mhs`
  show ?case
  proof (cases "f=g")
    case False with 6 show ?thesis by simp
  next
    case True
    with `mergeHTerms (HFun f fs # hsl) (HFun g gs # hsr) = Some mhs` 
    show ?thesis
    proof (cases "mergeHTerms fs gs", auto)
      fix m
      assume I: "(case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> Some (HFun g m # res)) = Some mhs"
      assume II: "mergeHTerms fs gs = Some m"
      show "P (HFun g fs # hsl) (HFun g gs # hsr) mhs" using I II
      proof (cases "mergeHTerms hsl hsr", auto)
	case (Some rres)
	with prems show "P (HFun g fs # hsl) (HFun g gs # hsr) (HFun g m # rres)" by auto
      qed
    qed
  qed
qed

lemma mergeHTermsSound : assumes "mergeHTerms hsl hsr = Some mhs" shows 
  "length hsl = length hsr \<and> length hsl = length mhs \<and> (\<forall> i < length hsl. [hsl ! i] \<inter> [hsr ! i] = [mhs ! i])"
proof (induct rule: mergeHTerms_induct)
  case EE
  thus ?case by simp
next
  case (BL h hsl hsr mhs)
  thus ?case proof (simp)
    show "\<forall> i < Suc (length hsr). [(\<box> # hsl) ! i] \<inter> [(h # hsr) ! i] = [(h # mhs) !i]"
    proof (rule allI, rule impI)
      fix i
      assume ilen: "i < Suc (length hsr)"
      show "[(\<box> # hsl) ! i] \<inter> [(h # hsr) ! i] = [(h # mhs) !i]" using BL and ilen by (cases i, auto)
    qed
  qed
next
  case (BR h hsl hsr mhs)
  thus ?case proof (simp)
    show "\<forall> i < Suc (length hsr). [(h # hsl) ! i] \<inter> [(\<box> # hsr) ! i] = [(h # mhs) ! i]"
    proof (rule allI, rule impI)
      fix i
      assume ilen: "i < Suc (length hsr)"
      show "[(h # hsl) ! i] \<inter> [(\<box> # hsr) ! i] = [(h # mhs) !i]" using BR and ilen by (cases i, auto)
    qed
  qed
next    
  case (FF f ls rs hsl hsr m mhs)
  thus ?case proof (simp)
    show "\<forall> i < Suc (length hsr). [(HFun f ls # hsl) ! i] \<inter> [(HFun f rs # hsr) ! i] = [(HFun f m # mhs) ! i]"
    proof (rule allI, rule impI)
      fix i
      assume ilen: "i < Suc (length hsr)"
      show "[(HFun f ls # hsl) ! i] \<inter> [(HFun f rs # hsr) ! i] = [(HFun f m # mhs) ! i]" 
      proof (cases i)
	case (Suc j)
	with ilen and FF have "[hsl ! j] \<inter> [hsr ! j] = [mhs ! j]" by simp
	with Suc show ?thesis by simp
      next
	case 0
	from FF have lr: "length ls = length rs" and lm: "length ls = length m" and mem: "\<forall> i < length ls. [ls ! i] \<inter> [rs ! i] = [m ! i]" by auto 
	have "[HFun f ls] \<inter> [HFun f rs] = [HFun f m]" 
	proof (rule set_ext, rule iffI)
	  fix t 	  
	  assume "t \<in> [HFun f m]" 
	  hence "\<exists> ts. t = Fun f ts \<and> length ts = length m \<and> (\<forall> k < length m. ts ! k \<in> [m ! k])" by simp
	  from this obtain ts where "t = Fun f ts" and "length ts = length m" and "\<forall> k < length m. ts ! k \<in> [m ! k]" by force
	  with lr lm mem show "t \<in> [HFun f ls] \<inter> [HFun f rs]" by auto
	next
	  fix t 
	  assume elem: "t \<in> [HFun f ls] \<inter> [HFun f rs]"
	  hence "\<exists> tsl. t = Fun f tsl \<and> length tsl = length ls \<and> (\<forall> k < length ls. tsl ! k \<in> [ls ! k])" by simp
	  from this obtain tsl where l1: "t = Fun f tsl" and l2: "length tsl = length ls" and l3: "\<forall> k < length ls. tsl ! k \<in> [ls ! k]" by force
	  from elem have "\<exists> tsr. t = Fun f tsr \<and> length tsr = length rs \<and> (\<forall> k < length rs. tsr ! k \<in> [rs ! k])" by simp
	  from this obtain tsr where r1: "t = Fun f tsr" and r3: "\<forall> k < length rs. tsr ! k \<in> [rs ! k]" by force
	  from l1 and r1  have id: "tsl = tsr" by simp
	  have "\<forall> k < length m. tsl ! k \<in> [m ! k]"
	  proof (rule allI, rule impI)
	    fix k 
	    assume klen: "k < length m"
	    with l3 lm have lmem: "tsl ! k \<in> [ls ! k]" by simp
	    from klen lr lm r3 id have rmem: "tsl ! k \<in> [rs ! k]" by simp
	    from lmem rmem klen mem lm show "tsl ! k \<in> [m ! k]" by force
	  qed
	  with l1 l2 lm show "t \<in> [HFun f m]" by simp
	qed
	with 0 show ?thesis by simp
      qed
    qed
  qed
next
  case Pre from assms show ?case .
qed   
	
fun merge :: "('f,'v) hterm \<Rightarrow> ('f,'v) hterm \<Rightarrow> ('f,'v) hterm option"
  where "merge hl hr = (case mergeHTerms [hl] [hr] of None \<Rightarrow> None | Some res \<Rightarrow> Some (res ! 0))"

lemma mergeComplete : assumes "merge hl hr = None" shows "[hl] \<inter> [hr] = {}"
using assms
proof (cases "mergeHTerms [hl] [hr]", auto)
  fix t
  assume n: "mergeHTerms [hl] [hr] = None" and l: "t \<in> [hl]" and r: "t \<in> [hr]"
  from n have "length [hl] \<noteq> length [hr] \<or> (\<exists> i < length [hl]. [[hl] ! i] \<inter> [[hr] ! i] \<subseteq> {})" 
    by (rule mergeHTermsComplete)
  with l r show False by force
qed

lemma mergeSound: assumes m: "merge hl hr = Some m" shows "[hl] \<inter> [hr] = [m]"
using assms
proof (cases "mergeHTerms [hl] [hr]", simp)
  fix res
  assume mt: "mergeHTerms [hl] [hr] = Some res" 
  with m have z: "m = res ! 0" by simp
  from mt have "length [hl] = length [hr] \<and> length [hl] = length res \<and> (\<forall> i < length [hl]. [[hl] ! i] \<inter> [[hr] ! i] = [res ! i])"
    by (rule mergeHTermsSound)
  with z show "[hl] \<inter> [hr] = [m]" by simp
qed
  
  

function matchList :: "(('f,'v) hterm \<times> ('f,'v) term) list \<Rightarrow> (('f,'v) hterm \<times> 'v) list option"
  where "matchList [] = Some []"
    |   "matchList ((\<box>, _) # ps) = matchList ps"
    |   "matchList ((HFun f hs,Fun g ss) # ps) = (if f = g then 
             (case decomp hs ss of Some pairs \<Rightarrow> matchList (pairs @ ps) | None \<Rightarrow> None) else None)"
    |   "matchList ((HFun f hs, Var x) # ps) = 
             (case matchList ps of Some pairs \<Rightarrow> Some ((HFun f hs,x) # pairs) | None \<Rightarrow> None)"
by pat_completeness auto
termination 
proof -
  let ?M = "measures [\<lambda>l. listsum (map (sz \<circ> snd) l), length]"
  show ?thesis 
  proof
    show "wf ?M" by auto    
  next
    fix hs ps 
    show "(ps, (\<box>, hs) # ps) \<in> ?M" by auto
  next
    fix f hs x ps 
    show "(ps, (HFun f hs, Var x) # ps) \<in> ?M" by auto
  next
    fix f :: 'f and hs :: "('f,'v) hterm list" and g :: 'f and ss :: "('f,'v) term list" and ps res
    assume dec: "decomp hs ss = Some res"
    show "(res @ ps, (HFun f hs, Fun g ss) # ps) \<in> ?M" 
    proof (rule measures_less, auto)
      from dec have len: "length hs = length ss" by (rule decompLength)
      from dec have res: "res = zip hs ss" by (rule decompZip)
      from len and res show "listsum (map (sz \<circ> snd) res) < Suc (listsum (map sz ss))" 
	by (auto simp: map_compose)
    qed
  qed
qed

definition matchL :: "(('f,'v) hterm \<times> ('f,'v) term) set \<Rightarrow> ('f,'v) sub set"
   where "matchL ps == {\<sigma>. (\<forall> p \<in> ps. snd p \<cdot> \<sigma> \<in> [fst p])}"
abbreviation matchLi :: "(('f,'v) hterm \<times> 'v) set \<Rightarrow> ('f,'v) sub set"
    where "matchLi ps == matchL ((\<lambda> p. (fst p, Var (snd p))) ` ps)"

lemma matchLUnion[simp] : "matchL (ps \<union> qs) = matchL ps \<inter> matchL qs"
  by (unfold matchL_def, blast)

lemma matchLInsert : "matchL (insert p ps) = {\<sigma>. snd p \<cdot> \<sigma> \<in> [fst p]} \<inter> matchL ps"
  by (unfold matchL_def, blast)

lemma matchLZip : assumes "length ts = length hs" shows 
  "{\<sigma>. \<forall>i<length ts. ts ! i \<cdot> \<sigma> \<in> [hs ! i]} = matchL (set (zip hs ts))" 
using assms
proof (induct ts arbitrary:hs) 
  case Nil thus ?case by (cases hs, auto simp:matchL_def)
next
  case (Cons t ss)
  thus ?case 
  proof (cases hs)
    case Nil with Cons show ?thesis by simp
  next
    case (Cons h hhs)
    from prems have "length ss = length hhs" by simp
    with prems have Prop: "{\<sigma>. \<forall>i<length ss. ss ! i \<cdot> \<sigma> \<in> [hhs ! i]} = matchL (set (zip hhs ss))" by simp
    have "{\<sigma>. \<forall>i<length (t # ss). (t # ss) ! i \<cdot> \<sigma> \<in> [(h # hhs) ! i]} = { \<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> {\<sigma>. \<forall>i<length ss. (ss ! i) \<cdot> \<sigma> \<in> [hhs ! i]}"
    proof (rule set_ext, rule iffI)
      fix s
      assume "s \<in> {\<sigma>. \<forall>i<length (t # ss). (t # ss) ! i \<cdot> \<sigma> \<in> [(h # hhs) ! i]}"
      thus "s  \<in> {\<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> {\<sigma>. \<forall>i<length ss. ss ! i \<cdot> \<sigma> \<in> [hhs ! i]}" by force  
    next
      fix s 
      assume Prop2: "s \<in>  {\<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> {\<sigma>. \<forall>i<length ss. ss ! i \<cdot> \<sigma> \<in> [hhs ! i]}"
      hence one: "t \<cdot> s \<in> [h]" by simp
      from Prop2 have two: "\<forall>i< length ss. ss ! i \<cdot> s \<in> [hhs ! i]" by simp
      have "\<forall> i < length (t # ss). (t # ss) ! i \<cdot> s \<in> [(h # hhs) ! i]"
      proof (rule allI, rule impI)
	fix i
	assume "i < length (t # ss)"
	with one two show "(t # ss) ! i \<cdot> s \<in> [(h # hhs) ! i]"
	  by (cases i, auto)
      qed
      thus "s \<in> {\<sigma>. \<forall>i<length (t # ss). (t # ss) ! i \<cdot> \<sigma> \<in> [(h # hhs) ! i]}" by simp
    qed
    also have "... =  { \<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> matchL (set (zip hhs ss))" using Prop by simp
    finally show ?thesis using Cons by (simp add: matchLInsert)
  qed
qed

lemma matchLFun[simp] : "matchL (insert (HFun f hs, Fun g ts) ps) = (if f = g \<and> length hs = length ts then matchL (set (zip hs ts)) \<inter> matchL ps else
{})" 
proof (cases "f = g \<and> length hs = length ts")
  assume "\<not> (f = g \<and> length hs = length ts)"
  hence "f \<noteq> g \<or> length hs \<noteq> length ts" by blast
  thus ?thesis
  proof
    assume "f \<noteq> g" thus ?thesis by (unfold matchL_def, simp)
  next
    assume "length hs \<noteq> length ts" thus ?thesis by (unfold matchL_def, auto)
  qed
next
  assume P: "f = g \<and> length hs = length ts"
  hence "{\<sigma>. \<forall>i<length ts. ts ! i \<cdot> \<sigma> \<in> [hs ! i]} = matchL (set (zip hs ts))" by (simp add: matchLZip)
  with P show ?thesis by (simp add: matchLInsert)
qed

    

lemma matchListComplete : assumes "matchList ps = None" shows "matchL (set ps) = {}"
using assms 
proof (induct rule:matchList.induct)
  case 1 thus ?case by simp
next
  case (2 t ps) thus ?case by (unfold matchL_def, auto)
next
  case (4 f hs x ps) thus ?case by (cases "matchList ps", unfold matchL_def, auto)
next
  case (3 f hs g ss ps) thus ?case 
  proof (cases "f = g")
    case False thus ?thesis by (unfold matchL_def, auto)
  next
    case True thus ?thesis 
    proof (cases "decomp hs ss")
      case (Some pairs)
      hence len: "length hs = length ss" by simp
      from Some have zip: "pairs = zip hs ss" by simp
      with Some `f = g` 3 have "matchL (set (zip hs ss @ ps)) = {}"  by simp
      with len show ?thesis by simp
    next
      case None
      hence difflen: "length hs \<noteq> length ss" by (rule decompNone)
      thus ?thesis by simp
    qed
  qed
qed

lemma matchListSound : assumes "matchList ps = Some qs" shows "matchL (set ps) = matchLi (set qs)" 
using assms 
proof (induct arbitrary:qs rule:matchList.induct)
  case 1 thus ?case by (unfold matchL_def, simp)
next
  case (2 t ps) thus ?case by (unfold matchL_def, auto)
next
  case (3 f hs g ss ps) thus ?case 
  proof (cases "f = g")
    case False with 3 show ?thesis by auto
  next
    case True thus ?thesis 
    proof (cases "decomp hs ss")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some pairs)
      hence len: "length hs = length ss" by simp
      from Some have zip: "pairs = zip hs ss" by simp
      with 3 Some True zip have "matchL (set (zip hs ss @ ps)) = matchLi (set qs)" by simp
      with len True show ?thesis by simp
    qed
  qed
next
  case (4 f hs x ps) thus ?case 
  proof (cases "matchList ps")
    case None with 4 show ?thesis by simp
  next
    case (Some pairs) with 4 show ?thesis by (auto simp: matchLInsert)
  qed
qed

fun mergeX :: "'v \<Rightarrow> ('f,'v) hterm \<Rightarrow> (('f,'v) hterm \<times> 'v) list \<Rightarrow> (('f,'v) hterm \<times> 'v) list option"
  where "mergeX x h [] = Some []"
      | "mergeX x h ((ht,y) # ps) = (if x = y then (case merge h ht of None \<Rightarrow> None | Some m \<Rightarrow> mergeX x m ps) 
                     else (case mergeX x h ps of None \<Rightarrow> None | Some pairs \<Rightarrow> Some ((ht,y) # pairs)))"

lemma mergeXComplete : assumes "mergeX x h ps = None"
  shows "matchLi (insert (h,x) (set ps)) = {}"
using assms
proof (induct ps arbitrary: h)
  case Nil thus ?case by simp
next
  case (Cons p ps)
  show ?case 
  proof (cases p)
    case (Pair ht y)
    show ?thesis 
    proof (cases "x = y")
      case False
      from Cons Pair False have "mergeX x h ps = None" by (cases "mergeX x h ps", auto)
      with Cons have "matchLi (insert (h, x) (set ps)) = {}" by simp
      thus ?thesis by (auto simp: matchLInsert)
    next
      case True
      show ?thesis
      proof (cases "merge h ht")
	case None
	hence "[h] \<inter> [ht] = {}" by (rule mergeComplete)
	with True Pair show ?thesis by (auto simp: matchLInsert)
      next
	case (Some m)
	with Pair True Cons have ind: "matchLi (insert (m,x) (set ps)) = {}" by simp
	from Some have "[h] \<inter> [ht] = [m]" by (rule mergeSound)
	with Pair ind True show ?thesis by (auto simp:matchLInsert)
      qed
    qed
  qed
qed

lemma mergeXSound : assumes "mergeX x h ps = Some rps"
  shows "\<exists> rh. (matchLi (insert (h,x) (set ps)) = matchLi (insert (rh,x) (set rps)) \<and> x \<notin> snd ` set rps)"
using assms
proof (induct ps arbitrary: h rps)
  case Nil thus ?case by auto
next
  case (Cons p ps)
  show ?case 
  proof (cases p)
    case (Pair ht y)
    show ?thesis 
    proof (cases "x = y")
      case False
      show ?thesis 
      proof (cases "mergeX x h ps")
	case None with Cons Pair False show ?thesis by simp
      next
	case (Some pairs)
	with Cons Pair False Some 
	obtain m where "matchLi (insert (h,x) (set ps)) = matchLi (insert (m,x) (set pairs)) \<and> x \<notin> snd ` set pairs" 
          and "rps = (ht,y) # pairs" by force 
	with False Pair show ?thesis by (auto simp: matchLInsert)
      qed
    next
      case True 
      show ?thesis
      proof (cases "merge h ht")
	case None with Pair Cons True show ?thesis by simp
      next
	case (Some m)
	with Pair True Cons obtain rh where 
	  ind: "matchLi (insert (m,x) (set ps)) = matchLi (insert (rh,x) (set rps)) \<and> x \<notin> snd ` set rps" by force
	from Some have "[h] \<inter> [ht] = [m]" by (rule mergeSound)
	with Pair ind True show ?thesis by (auto simp:matchLInsert)
      qed
    qed
  qed
qed

lemma mergeXSize[simp] : assumes "mergeX x h ps = Some rps" shows "length ps \<ge> length rps"
using assms
proof (induct ps arbitrary: h rps)
  case Nil thus ?case by simp
next
  case (Cons p ps)
  show ?case
  proof (cases p)
    case (Pair ht y)
    show ?thesis
    proof (cases "x = y")
      case False
      with Cons Pair show ?thesis by (cases "mergeX x h ps", auto)
    next
      case True
      with Cons Pair show ?thesis by (cases "merge h ht", auto, force)
    qed
  qed
qed  

fun mergeAll :: "(('f,'v) hterm \<times> 'v) list \<Rightarrow> bool"
  where "mergeAll [] = True"
     |  "mergeAll ((h,x) # ps) = (case mergeX x h ps of None \<Rightarrow> False | Some mps \<Rightarrow> mergeAll mps)"

lemma mergeAllComplete : fixes ps :: "(('f,'v) hterm \<times> 'v) list"
  shows "\<not> mergeAll ps \<Longrightarrow> matchLi (set ps) = {}"
proof (induct rule: length_induct)
  fix pps :: "(('f,'v) hterm \<times> 'v) list"
  assume ind: "\<forall> mps. length (mps :: (('f,'v) hterm \<times> 'v) list) < length pps \<longrightarrow> \<not> mergeAll mps \<longrightarrow> matchLi (set mps) = {}" and
    res: "\<not> mergeAll pps"
  show "matchLi (set pps) = {}"
  proof (cases pps)
    case Nil with res show ?thesis by simp
  next
    case (Cons p ps)
    show ?thesis
    proof (cases p)
      case (Pair h x)
      show ?thesis
      proof (cases "mergeX x h ps")
	case None
	hence "matchLi (insert (h,x) (set ps)) = {}" by (rule mergeXComplete)
	with Cons Pair show ?thesis by simp
      next
	case (Some mps)
	hence "\<exists> rh. (matchLi (insert (h,x) (set ps)) = matchLi (insert (rh,x) (set mps)) \<and> x \<notin> snd ` set mps)" by (rule mergeXSound)
	from this obtain rh where m: "matchLi (insert (h,x) (set ps)) = matchLi (insert (rh,x) (set mps))" by auto
	from Some have "length mps \<le> length ps" by simp
	with Cons have len: "length mps < length pps" by simp
	from Cons Pair Some res have "\<not> mergeAll mps" by auto
	with len ind have "matchLi (set mps) = {}" by auto
	with m Pair Cons show ?thesis by (simp add: matchLInsert)
      qed
    qed
  qed
qed

lemma mergeAllSound : fixes ps :: "(('f,'v) hterm \<times> 'v) list"
  shows "mergeAll ps \<Longrightarrow> \<exists> \<sigma>. \<sigma> \<in> matchLi (set ps)"
proof (induct rule: length_induct)
  fix pps :: "(('f,'v) hterm \<times> 'v) list"
  assume ind: "\<forall> mps. length (mps :: (('f,'v) hterm \<times> 'v) list) < length pps \<longrightarrow> mergeAll mps \<longrightarrow> (\<exists> \<sigma>. \<sigma> \<in> matchLi (set mps))" and
    res: "mergeAll pps"
  show "\<exists> \<sigma> . \<sigma> \<in> matchLi (set pps)"
  proof (cases pps)
    case Nil with res show ?thesis by (unfold matchL_def, simp)
  next
    case (Cons p ps)
    show ?thesis
    proof (cases p)
      case (Pair h x)
      show ?thesis
      proof (cases "mergeX x h ps")
	case None with Cons Pair res show ?thesis by simp
      next
	case (Some mps)
	hence "\<exists> rh. (matchLi (insert (h,x) (set ps)) = matchLi (insert (rh,x) (set mps)) \<and> x \<notin> snd ` set mps)" by (rule mergeXSound)
	from this obtain rh where m: "matchLi (insert (h,x) (set ps)) = matchLi (insert (rh,x) (set mps))" 
	  and noX: "x \<notin> snd ` set mps" by auto
	from Some have "length mps \<le> length ps" by simp
	with Cons have len: "length mps < length pps" by simp
	from Cons Pair Some res have "mergeAll mps" by auto
	with len ind obtain \<sigma> where sigma: "\<sigma> \<in> matchLi (set mps)" by auto
	show ?thesis 
	proof (cases \<sigma>)
	  case (Sub sigma)
	  have "\<exists> t. t \<in> [rh]" by (rule equiv_classes_nonempty)
	  from this obtain t where trh: "t \<in> [rh]" ..
	  let ?delta = "Sub (\<lambda> y. if x = y then t else sigma y)"
	  have "?delta \<in> matchLi (insert (rh, x) (set mps))" using trh 
	  proof (simp add: matchLInsert)
	    show "?delta \<in> matchLi (set mps)" using Sub sigma noX by (auto simp: matchL_def)
	  qed
	  with m Pair Cons show ?thesis by auto
	qed
      qed
    qed
  qed
qed

fun matchI :: "('f,'v) hterm \<Rightarrow> ('f,'v) term \<Rightarrow> bool"
  where "matchI h t = (case matchList [(h,t)] of None \<Rightarrow> False | Some pairs \<Rightarrow> mergeAll pairs)"

lemma matchISound : "match h t = matchI h t" 
proof (cases "matchList [(h,t)]")
  assume mln: "matchList [(h,t)] = None"
  hence "matchL (set [(h,t)]) = {}" by (rule matchListComplete)
  with mln show ?thesis by (simp add: matchL_def match_def)
next
  fix pairs
  assume mls: "matchList [(h,t)] = Some pairs"
  hence one: "matchL (set [(h,t)]) = matchLi (set pairs)" by (rule matchListSound)
  show ?thesis
  proof (cases "mergeAll pairs")
    case False
    hence "matchLi (set pairs) = {}" by (rule mergeAllComplete)
    with one mls have "matchL {(h,t)} = {}" by simp
    hence "\<not> match h t" by (simp add: matchL_def match_def)
    with mls one False show ?thesis by simp
  next
    case True
    hence "\<exists> \<sigma>. \<sigma> \<in> matchLi (set pairs)" by (rule mergeAllSound)
    with one mls have "\<exists> \<sigma> . \<sigma> \<in> matchL {(h,t)}" by auto
    hence "match h t" by (simp add: matchL_def match_def)
    with mls one True show ?thesis by simp
  qed
qed
    
  

(* a semantic version of cap which is not executable *)

fun cap :: "('f,'v) rule set \<Rightarrow> ('f,'v) term  \<Rightarrow> ('f,'v) hterm"
  where "cap _ (Var _) = \<box>"
    |   "cap R (Fun f ts) = 
            (let h = HFun f (map (cap R) ts) in (
                if (\<exists> r \<in> R. match h (fst r)) then \<box> else h
            ))"

(* an executable version of cap *)
fun capI :: "('f,'v) rule list \<Rightarrow> ('f,'v) term  \<Rightarrow> ('f,'v) hterm"
  where "capI _ (Var _) = \<box>"
    |   "capI R (Fun f ts) = 
            (let h = HFun f (map (capI R) ts) in (
                if testPos (\<lambda> r. matchI h (fst r)) R then \<box> else h
            ))"

(* and a version for the reversed TRS *)
fun capR :: "('f,'v) rule list \<Rightarrow> ('f,'v) term  \<Rightarrow> ('f,'v) hterm"
  where "capR _ (Var _) = \<box>"
    |   "capR R (Fun f ts) = 
            (let h = HFun f (map (capR R) ts) in (
                if testPos (\<lambda> r. matchI h (snd r)) R  then \<box> else h
            ))"


lemma capISound :   fixes R :: "('f,'v) rule list"
  and t :: "('f,'v) term"
  shows "capI R t = cap (set R) t"
proof (induct t)
    case (Var x) thus ?case by simp
  next
    case (Fun f ts)
    let ?h = "HFun f (map (cap (set R)) ts) :: ('f,'v) hterm"
    let ?hI = "HFun f (map (capI R) ts) :: ('f,'v) hterm"
    from Fun have heq: "?hI = ?h" by simp
    let ?EQ = "\<lambda>(h1 :: ('f,'v) hterm) (h2 :: ('f,'v) hterm) . ((\<exists> r \<in> set R. match h1 (fst r)) = (testPos (\<lambda> r. matchI h2 (fst r)) R ))"
    have "!! h1 h2 . h1 = h2 \<Longrightarrow> ?EQ h1 h2 "
    proof -
      fix h1 h2 :: "('f,'v)hterm" 
      show "h1 = h2 \<Longrightarrow> ?EQ h1 h2" by (induct R, auto simp:matchISound)
    qed
    hence "?h = ?hI \<Longrightarrow> ?EQ ?h ?hI" .
    with heq have a: "?EQ ?h ?hI" by simp
    show ?case using if_weak_cong[OF a, where x = "hterm.Hole" and y = "HFun f (map (cap (set R)) ts)",symmetric] unfolding capI.simps cap.simps Let_def heq matchISound by force
qed

lemma nrrstep_seq_imp_arg_rstep_seq : 
  assumes "(s,t) \<in> (nrrstep R)^*"
  shows "root s = root t \<and> num_args s = num_args t \<and> (\<forall> i < num_args s. (s |_ <i>, t |_ <i> ) \<in> (rstep R)^*)"
using assms 
proof (induct)
  case base show ?case by simp
next
  case (step u t)
  hence root: "root s = root u" and num: "num_args s = num_args u" by auto
  from `(u,t) \<in> nrrstep R` have fu: "\<exists> f us ts. u = Fun f us \<and> t = Fun f ts" by (rule nrrstep_equiv_root)
  with root have eqroot: "root s = root t" by auto
  from step have arg: "\<forall> i < num_args s. (s |_ <i> , u |_ <i>) \<in> (rstep R)^*" by simp
  from `(u,t) \<in> nrrstep R` have 
    "\<exists>j<num_args u. num_args u = num_args t \<and> (u |_ <j>, t |_ <j>) \<in> rstep R \<and> (\<forall>k. j \<noteq> k \<longrightarrow> u |_ <k> = t |_ <k>)"
    by (rule nrrstep_imp_subt_rstep)
  from this obtain j where jnum: "j < num_args u" and num2: "num_args u = num_args t" 
    and utj: "(u |_ <j>, t |_ <j>) \<in> rstep R"
    and utk: "\<forall>k. j \<noteq> k \<longrightarrow> u |_ <k> = t |_ <k>" by force
  from num and num2 have num3 : "num_args s = num_args t" by simp
  have "\<forall> i < num_args s. (s |_ <i> , t |_ <i>) \<in> (rstep R)^*" 
  proof auto
    fix i
    assume inum: "i < num_args s"
    show "(s |_ <i> , t|_<i>) \<in> (rstep R)^*" 
    proof (cases "j = i")
      case False
      with utk have last: "u |_ <i> = t |_ <i>" by simp
      from inum arg have "(s|_<i>, u|_<i>) \<in> (rstep R)^*" by simp
      from this and last show ?thesis by simp 
   next
      case True
      with utj have last: "(u|_<i>,t|_<i>) \<in> rstep R" by simp
      from inum arg have "(s|_<i>, u|_<i>) \<in> (rstep R)^*" by simp
      from this and last show ?thesis by simp
    qed
  qed
  from this and num3 and eqroot show ?case by simp
qed

lemma firstStep: assumes LSR: "L = S \<union> R" and xyL: "(x,y) \<in> L^*"
  shows "(x,y) \<in> R^* \<or> (x,y) \<in> L^* O S O R^*"
proof (cases "(x,y) \<in> R^*")
  case True
  thus ?thesis by simp
next
  case False 
  let ?SR = "S \<union> R"
  from xyL and LSR have "(x,y) \<in> ?SR^*" by simp
  from this and False have "(x,y) \<in> ?SR^* O S O R^*" 
  proof (induct rule: rtrancl_induct)
    case base thus ?case by simp
  next
    case (step y z)
    thus ?case
    proof (cases "(x,y) \<in> R^*")
      case False with step have "(x,y) \<in> ?SR^* O S O R^*" by simp
      from this obtain u where xu: "(x,u) \<in> S O R^*" and uy: "(u,y) \<in> ?SR^*" by force
      from `(y,z) \<in> ?SR` have "(y,z) \<in> ?SR^*" by auto
      with uy have "(u,z) \<in> ?SR^*" by (rule rtrancl_trans)
      with xu show ?thesis by auto
    next
      case True 
      have "(y,z) \<in> S" 
      proof (rule ccontr)
	assume "(y,z) \<notin> S" with `(y,z) \<in> ?SR` have "(y,z) \<in> R" by auto
	with True  have "(x,z) \<in> R^*"  by auto
	with `(x,z) \<notin> R^*` show False ..
      qed
      with True show ?thesis by auto
    qed
  qed
  with LSR show ?thesis by simp
qed  
  


lemma capSound : assumes "(s \<cdot> \<sigma>,t) \<in> (rstep R)^*" shows 
  "t \<in> [cap R s]"
using assms
proof (induct s arbitrary: t rule: term_induct)
  case (Var x) thus ?case by simp
next
  case (Fun f ss)
  let ?h = "HFun f (map (cap R) ss)" 
  show ?case 
  proof (cases "cap R (Fun f ss)")
    case Hole
    thus ?thesis by simp
  next
    case (HFun g hs)
    hence res: "cap R (Fun f ss) = ?h"
      by (cases "\<exists> r \<in> R. match ?h (fst r)", auto simp: Let_def)
    have nrr: "!! u. (Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^* \<Longrightarrow> u \<in> [?h]" 
    proof -
      fix u
      assume "(Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^*"
      hence cond: "root (Fun f ss \<cdot> \<sigma>) = root u \<and> num_args (Fun f ss \<cdot> \<sigma>) = num_args u \<and> (\<forall> i < num_args (Fun f ss \<cdot> \<sigma>). (Fun f ss \<cdot> \<sigma> |_ <i>, u |_ <i> ) \<in> (rstep R)^*)"	
	by (rule nrrstep_seq_imp_arg_rstep_seq)
      hence "root (Fun f ss) = root u" by simp
      hence "\<exists> us. u = Fun f us" by (cases u, auto)
      from this obtain us where u: "u = Fun f us" by force
      with cond have len: "length ss = length us" by simp      
      hence "Fun f us \<in> [?h]" 
      proof (auto)
	fix i
	assume ilen: "i < length us"
	with len have "<i> \<in> poss (Fun f ss)" by simp
	hence "Fun f ss \<cdot> \<sigma>|_<i> = Fun f ss |_<i> \<cdot> \<sigma>" by simp
	with ilen len cond have one: "(Fun f ss |_<i> \<cdot> \<sigma>, u|_<i>) \<in> (rstep R)^*" by simp	
	from ilen len have "Fun f ss |_<i> \<in> set ss" by simp
	with one Fun have "u|_<i> \<in> [cap R (Fun f ss |_<i>)]" by simp
	with u show "us ! i \<in> [cap R (ss ! i)]" by simp
      qed
      with u show "u \<in> [?h]" by simp
    qed
    
    have "rstep R = rrstep R \<union> nrrstep R" by (rule rstep_iff_rrstep_or_nrrstep)
    from this and `(Fun f ss \<cdot> \<sigma>, t) \<in> (rstep R)^*`
    have  "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^* \<or> (Fun f ss \<cdot> \<sigma>, t) \<in> (rstep R)^* O (rrstep R) O (nrrstep R)^*" by (rule firstStep)
    thus ?thesis
    proof 
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^*"
      with nrr res show ?thesis by (auto)
    next
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> (rstep R)^* O rrstep R O (nrrstep R)^*"
      from this obtain u v where fu: "(Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^*" and uv: "(u,v) \<in> rrstep R" by force
      from fu nrr have uh: "u \<in> [?h]" by simp
      from uv have "\<exists> l r \<sigma>. (l,r) \<in> R \<and> l \<cdot> \<sigma> = u \<and> r \<cdot> \<sigma> = v" by (rule rrstep_imp_rule_subs)
      from this obtain l r \<sigma> where lrR: "(l, r) \<in> R" and match: "l \<cdot> \<sigma> = u" by force
      from uh and match  have "match ?h l" by (unfold match_def, auto)
      with lrR have "cap R (Fun f ss) = \<box>" by (simp add: Let_def, force)
      thus ?thesis by simp
    qed
  qed
qed

lemma matchCapSound : 
  assumes derivation: "(t \<cdot> \<delta>, s \<cdot> \<sigma>) \<in> (rstep R)^*"
  shows "match (cap R t) s"
proof -
  from derivation have "s \<cdot> \<sigma> \<in> [cap R t]" by (rule capSound)
  thus ?thesis by (unfold match_def, auto)
qed

lemma matchICapISound : 
  assumes derivation:  "(t \<cdot> \<delta>, s \<cdot> \<sigma>) \<in> (rstep (set R))^*"
  shows "matchI (capI R t) s"
proof -
  from derivation have "match (cap (set R) t) s" by (rule matchCapSound)
  hence "matchI (cap (set R) t) s" by (simp add: matchISound)
  thus "matchI (capI R t) s" by (simp add: capISound)
qed
  
lemma matchICapRSound :
  assumes derivation: "(t \<cdot> \<sigma>, s \<cdot> \<delta>) \<in> (rstep (set R))^*"
  shows "matchI (capR R s) t"
proof -
  let ?R1 = "map (\<lambda> lr. (snd lr, fst lr)) R"
  have rev: "set ?R1 = (set R)^-1" by (induct R, auto)
  have id: "((rstep (set R))^-1)^* = ((rstep (set R))^*)^-1" by (rule rtrancl_converse)
  from derivation have "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> ((rstep (set R))^*)^-1" by simp
  hence "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> (rstep ((set R)^-1))^*" using id by simp
  with rev have "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> (rstep (set ?R1))^*" by simp
  hence mapResult: "matchI (capI ?R1 s) t" by (rule matchICapISound)

  have "capR R s = capI ?R1 s"
  proof (induct s)
    case (Var x) thus ?case by simp
  next
    case (Fun f ss) 
    hence eq1: "map (capR R) ss = map (capI ?R1) ss" by simp
    have eq2: "\<And>h. testPos (\<lambda> r. matchI h (snd r)) R  = testPos (\<lambda> r. matchI h (fst r)) ?R1"
    proof -
      fix h
      show "testPos (\<lambda> r. matchI h (snd r)) R = testPos (\<lambda> r. matchI h (fst r)) ?R1"
	by (induct R, auto)
    qed
    have "capR R (Fun f ss) =  (let h = HFun f (map (capI ?R1) ss) in 
                (if testPos (\<lambda> r. matchI h (snd r) ) R then \<box> else h))" by (simp add: eq1)
    also have "... = capI ?R1 (Fun f ss)" using eq2 by (simp add: Let_def) 
    finally show ?case .
  qed

  with mapResult show ?thesis by simp
qed

(* now that ren and cap have been defined we come to the definition of the 
   estimated dependency graph *)

fun checkEdg :: "('f,'v) rule list \<Rightarrow> (('f,'v)rule \<times> ('f,'v)rule) \<Rightarrow> bool"
  where "checkEdg R ((s,t),(u,v)) = (matchI (capI R t) u   \<and>    matchI (capR R u) t)"


lemma checkEdgSound[simp] : assumes seq: "(t \<cdot> \<sigma>, u \<cdot> \<delta>) \<in> (rstep (set R))^*" shows 
  "checkEdg R ((s,t),(u,v))"
proof -
  from seq have forward: "matchI (capI R t) u" by (rule matchICapISound)
  from seq have backward: "matchI (capR R u) t" by (rule matchICapRSound)
  from forward backward show ?thesis by simp
qed

(*
fun edg :: "('f,'v)rule list \<Rightarrow> (('f,'v)rule \<times> ('f,'v)rule) list \<Rightarrow> (('f,'v)rule \<times> ('f,'v)rule) list \<Rightarrow> bool"
  where "edg R [] rG  = True"
     |  "edg R (((s,t),(u,v)) # G) rG = (edg R G rG \<and> 
              (  ((s,t),(u,v)) mem rG   \<or> \<not> edgCheck R ((s,t),(u,v))))"


definition initEdg :: "('f,'v)rule list \<Rightarrow> ('f,'v)rule list \<Rightarrow> (('f,'v)rule \<times> ('f,'v)rule) list \<Rightarrow> bool"
  where "initEdg R P rG  = edg R [(st,uv) . st \<leftarrow> P, uv \<leftarrow> P] rG"


lemma edgSound : assumes edg: "edg R G rG"
  and fin: "finite_gdpp (graph ns rG, set R)" 
  shows "finite_gdpp (graph ns G, set R)"
proof -
  have main: "\<And> s t \<sigma>. min_graph_ichain (graph ns G, set R) s t \<sigma> \<Longrightarrow> min_graph_ichain (graph ns rG, set R) s t \<sigma>" 
  proof -
    fix s t \<sigma> 
    show "min_graph_ichain (graph ns G, set R) s t \<sigma> \<Longrightarrow> min_graph_ichain (graph ns rG, set R) s t \<sigma>"
    proof (auto simp: edges_def graph_def nodes_def)
      fix i
      assume "\<forall>i. ((s i, t i), s (Suc i), t (Suc i)) \<in> set G"
       and  "\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep (set R))^*"
      hence G: "((s i, t i), s (Suc i), t (Suc i)) \<in> set G" (is "((?s,?t),(?u,?v)) \<in> set G") 
	and step: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep (set R))^*" by auto      
      from G edg show "((?s,?t), (?u,?v)) \<in> set rG" (is "?e \<in> _")
      proof (induct G)
	case Nil thus ?case by simp
      next
	case (Cons edge GG)
	have "\<exists> ss tt uu vv. edge = ((ss,tt),(uu,vv))" by (cases edge, auto)
	from this obtain ss tt uu vv where id: "edge = ((ss,tt),(uu,vv))" (is "_ = ?edge") by blast
	with Cons have edgG: "edg R (?edge # GG) rG" by blast
	hence edg: "edg R GG rG" by simp
	show ?case
	proof (cases "?e \<in> set GG")
	  case True
	  with Cons edg show ?thesis by simp
	next
	  case False
	  with Cons and id have "?e = ?edge" by simp 
	  with Cons id have edgSuccess: "edg R (?e # GG) rG" by blast
	  from step have "edgCheck R ((?s,?t),(?u,?v))" by (rule edgCheckSound)
	  from edgSuccess this  show ?thesis 	    
	    by (auto, cases "?e mem rG", auto simp:mem_iff)
	qed
      qed
    qed
  qed
  with fin show ?thesis by (unfold finite_gdpp_def, blast)
qed



lemma initEdgSound : assumes initEdg: "initEdg R P rG"
  and fin: "finite_gdpp (graph P rG, set R)"
  shows "finite_dpp (set P, set R)"
proof -
  have part1: "\<And> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<Longrightarrow> min_graph_ichain (graph P [(st,uv). st \<leftarrow> P, uv \<leftarrow> P], set R) s t \<sigma>"
    by (auto simp: edges_def graph_def nodes_def)
  from initEdg fin have "finite_gdpp (graph P [(st,uv). st \<leftarrow> P, uv \<leftarrow> P], set R)" by (simp add: initEdg_def edgSound)
  with part1 show ?thesis by (simp add: finite_dpp_def finite_gdpp_def del: min_ichain.simps min_graph_ichain.simps, blast) 
qed
*)

declare checkEdg.simps[simp del]


end
