(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Show imports Main (* Code_Message *)
begin

types sstring = string

class "show" = type + 
  fixes "show" :: "'a \<Rightarrow> string"

definition unquote :: "sstring \<Rightarrow> sstring"
  where "unquote x \<equiv> filter (\<lambda> c. \<not> c = CHR ''\"'') x"
definition showF :: "'a :: show \<Rightarrow> sstring"
  where "showF x \<equiv> unquote (show x)"

class sshow = type +
  fixes sshow :: "'a \<Rightarrow> sstring"


abbreviation sAppend :: "sstring \<Rightarrow> sstring \<Rightarrow> sstring" (infixl "+++" 10)
  where "sAppend s t \<equiv> (s @ t)"


abbreviation sConcat :: "sstring list \<Rightarrow> sstring" where "sConcat \<equiv> concat"


definition sEmpty :: "sstring" where "sEmpty = []"


instantiation bool  :: sshow
begin
fun sshow_bool where "sshow_bool True = ''True''" | "sshow_bool False = ''False''"
instance by intro_classes
end

instantiation char  :: sshow
begin
fun sshow_char where "sshow_char x = ([x] :: string)"
instance by intro_classes
end

instantiation list :: ("sshow") "sshow"
begin
fun sshow_list where "sshow xs = sConcat (map sshow xs)"
instance by intro_classes
end

definition newline :: string where "newline \<equiv> '' newline ''"


instantiation * :: (sshow,sshow) sshow
begin
fun sshow_prod where "sshow_prod (x,y) = (concat [sshow x, '' -> '', sshow y])"
instance by intro_classes
end


fun showList  :: "sstring list \<Rightarrow> string"
 where "showList xs = concat (map (\<lambda> x. x @ newline) xs)"


instantiation nat :: sshow
begin
  fun sshow_nat where "sshow_nat 0 = ''0''" | 
        "sshow_nat (Suc n) = ( ''S('' +++  sshow_nat n +++ '')'')"
instance by intro_classes
end


instantiation int :: sshow
begin
function int_to_abs :: "int \<Rightarrow> nat" where "int_to_abs i = (if i \<le> 0 then 0 else Suc (int_to_abs (i - 1)))"
by (pat_completeness, auto)
termination proof
  show "wf (int_ge_less_than (0::int))" by (rule wf_int_ge_less_than)
next
  fix i :: int assume "\<not> i \<le> 0" thus "(i - 1, i) \<in> int_ge_less_than 0" 
    unfolding int_ge_less_than_def by auto
qed
fun sshow_int where "sshow_int i = (if i \<ge> 0 then sshow (int_to_abs i) else (''-'' +++  sshow (int_to_abs (0 - i))))"
instance by intro_classes
end 


end


